﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.ScatterSearch {
  /// <summary>
  /// An operator that updates the reference set.
  /// </summary>
  [Item("ReferenceSetUpdateMethod", "An operator that updates the reference set.")]
  [StorableClass]
  public sealed class ReferenceSetUpdateMethod : SingleSuccessorOperator {
    #region Parameter properties
    private ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public ValueLookupParameter<IntValue> NumberOfHighQualitySolutionsParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["NumberOfHighQualitySolutions"]; }
    }
    public ValueLookupParameter<IntValue> ReferenceSetSizeParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["ReferenceSetSize"]; }
    }
    #endregion

    #region Properties
    public IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    public IntValue NumberOfHighQualitySolutions {
      get { return NumberOfHighQualitySolutionsParameter.ActualValue; }
      set { NumberOfHighQualitySolutionsParameter.ActualValue = value; }
    }
    public IntValue ReferenceSetSize {
      get { return ReferenceSetSizeParameter.ActualValue; }
      set { ReferenceSetSizeParameter.ActualValue = value; }
    }
    #endregion

    [StorableConstructor]
    private ReferenceSetUpdateMethod(bool deserializing) : base(deserializing) { }
    private ReferenceSetUpdateMethod(ReferenceSetUpdateMethod original, Cloner cloner) : base(original, cloner) { }
    public ReferenceSetUpdateMethod() : base() { Initialize(); }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ReferenceSetUpdateMethod(this, cloner);
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new ScopeParameter("CurrentScope", "The current scope to which the new solutions are added as sub-scopes."));
      Parameters.Add(new ValueLookupParameter<IntValue>("NumberOfHighQualitySolutions", "The number of high quality solutions that should be added to the reference set."));
      Parameters.Add(new ValueLookupParameter<IntValue>("ReferenceSetSize", "The size of the reference set."));
      #endregion

      #region Create operators
      #endregion

      #region Create operator graph
      #endregion
    }

    public override IOperation Apply() {
      for (int i = 0; i < ReferenceSetSize.Value - NumberOfHighQualitySolutions.Value; i++) {
        IScope mostDiverseSolution = null;
        int maxDiversity = 0;
        foreach (var p in CurrentScope.SubScopes[0].SubScopes) {
          int diversity = 0;
          BinaryVector pSol = p.Variables["KnapsackSolution"].Value as BinaryVector;
          foreach (var r in CurrentScope.SubScopes[1].SubScopes) {
            BinaryVector rSol = r.Variables["KnapsackSolution"].Value as BinaryVector;
            if (pSol != null && rSol != null) diversity += CalculateDiversity(pSol, rSol);
          }
          if (mostDiverseSolution == null || diversity > maxDiversity) {
            mostDiverseSolution = p;
            maxDiversity = diversity;
          }
        }
        CurrentScope.SubScopes[1].SubScopes.Add(mostDiverseSolution); // Update parent?
        CurrentScope.SubScopes[0].SubScopes.Remove(mostDiverseSolution); // Update parent?
      }
      return base.Apply();
    }

    private int CalculateDiversity(BinaryVector pSol, BinaryVector rSol) {
      int diversity = 0;
      for (int i = 0; i < pSol.Length; i++)
        if (pSol[i] != rSol[i]) diversity++;
      return diversity;
    }
  }
}
