﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Algorithms.LocalSearch;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.ScatterSearch {
  /// <summary>
  /// An operator which represents a scatter search.
  /// </summary>
  [Item("ScatterSearchMainLoop", "An operator which represents the main loop of a scatter search.")]
  [StorableClass]
  public sealed class ScatterSearchMainLoop : AlgorithmOperator {
    #region Parameter properties
    internal LookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }
    internal LookupParameter<IntValue> IterationsParameter {
      get { return (LookupParameter<IntValue>)Parameters["Iterations"]; }
    }
    internal LookupParameter<IntValue> MaximumIterationsParameter {
      get { return (LookupParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    internal LookupParameter<BoolValue> MaximizationParameter {
      get { return (LookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    internal LookupParameter<IntValue> SampleSizeParameter {
      get { return (LookupParameter<IntValue>)Parameters["SampleSize"]; }
    }
    internal LookupParameter<IntValue> PopulationSizeParameter {
      get { return (LookupParameter<IntValue>)Parameters["PopulationSize"]; }
    }
    internal LookupParameter<IntValue> ReferenceSetSizeParameter {
      get { return (LookupParameter<IntValue>)Parameters["ReferenceSetSize"]; }
    }
    internal LookupParameter<IntValue> NumberOfHighQualitySolutionsParameter {
      get { return (LookupParameter<IntValue>)Parameters["NumberOfHighQualitySolutions"]; }
    }
    internal LookupParameter<DoubleValue> QualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    internal LookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    internal LookupParameter<DoubleValue> MoveQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    internal LookupParameter<VariableCollection> ResultsParameter {
      get { return (LookupParameter<VariableCollection>)Parameters["Results"]; }
    }
    internal LookupParameter<IOperator> MoveEvaluatorParameter {
      get { return (LookupParameter<IOperator>)Parameters["MoveEvaluator"]; }
    }
    internal LookupParameter<IOperator> MoveGeneratorParameter {
      get { return (LookupParameter<IOperator>)Parameters["MoveGenerator"]; }
    }
    internal LookupParameter<IOperator> MoveMakerParameter {
      get { return (LookupParameter<IOperator>)Parameters["MoveMaker"]; }
    }
    internal LookupParameter<IMultiAnalyzer> AnalyzerParameter {
      get { return (LookupParameter<IMultiAnalyzer>)Parameters["Analyzer"]; }
    }
    #endregion

    #region Properties
    public IRandom Random {
      get { return RandomParameter.ActualValue; }
      set { RandomParameter.ActualValue = value; }
    }
    public IntValue Iterations {
      get { return IterationsParameter.ActualValue; }
      set { IterationsParameter.ActualValue = value; }
    }
    public IntValue MaximumIterations {
      get { return MaximumIterationsParameter.ActualValue; }
      set { MaximumIterationsParameter.ActualValue = value; }
    }
    public BoolValue Maximization {
      get { return MaximizationParameter.ActualValue; }
      set { MaximizationParameter.ActualValue = value; }
    }
    public IntValue SampleSize {
      get { return SampleSizeParameter.ActualValue; }
      set { SampleSizeParameter.ActualValue = value; }
    }
    public IntValue PopulationSize {
      get { return PopulationSizeParameter.ActualValue; }
      set { PopulationSizeParameter.ActualValue = value; }
    }
    public IntValue ReferenceSetSize {
      get { return ReferenceSetSizeParameter.ActualValue; }
      set { ReferenceSetSizeParameter.ActualValue = value; }
    }
    public IntValue NumberOfHighQualitySolutions {
      get { return NumberOfHighQualitySolutionsParameter.ActualValue; }
      set { NumberOfHighQualitySolutionsParameter.ActualValue = value; }
    }
    public DoubleValue Quality {
      get { return QualityParameter.ActualValue; }
      set { QualityParameter.ActualValue = value; }
    }
    public DoubleValue BestKnownQuality {
      get { return BestKnownQualityParameter.ActualValue; }
      set { BestKnownQualityParameter.ActualValue = value; }
    }
    public DoubleValue MoveQuality {
      get { return MoveQualityParameter.ActualValue; }
      set { MoveQualityParameter.ActualValue = value; }
    }
    public VariableCollection Results {
      get { return ResultsParameter.ActualValue; }
      set { ResultsParameter.ActualValue = value; }
    }
    public IOperator MoveEvaluator {
      get { return MoveEvaluatorParameter.ActualValue; }
      set { MoveEvaluatorParameter.ActualValue = value; }
    }
    public IOperator MoveGenerator {
      get { return MoveGeneratorParameter.ActualValue; }
      set { MoveGeneratorParameter.ActualValue = value; }
    }
    public IOperator MoveMaker {
      get { return MoveMakerParameter.ActualValue; }
      set { MoveMakerParameter.ActualValue = value; }
    }
    public IMultiAnalyzer Analyzer {
      get { return AnalyzerParameter.ActualValue; }
      set { AnalyzerParameter.ActualValue = value; }
    }
    #endregion

    [StorableConstructor]
    private ScatterSearchMainLoop(bool deserializing) : base(deserializing) { }
    private ScatterSearchMainLoop(ScatterSearchMainLoop original, Cloner cloner) : base(original, cloner) { }
    public ScatterSearchMainLoop() : base() { Initialize(); }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ScatterSearchMainLoop(this, cloner);
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new LookupParameter<IRandom>("Random", "A pseudo random number generator."));
      Parameters.Add(new LookupParameter<IntValue>("Iterations", "The number of iterations performed."));
      Parameters.Add(new LookupParameter<IntValue>("MaximumIterations", "The maximum number of generations which should be processed."));
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, otherwise false."));
      Parameters.Add(new LookupParameter<IntValue>("SampleSize", "Number of moves that MultiMoveGenerators should create. This is ignored for Exhaustive- and SingleMoveGenerators."));
      Parameters.Add(new LookupParameter<IntValue>("PopulationSize", "The size of the population."));
      Parameters.Add(new LookupParameter<IntValue>("ReferenceSetSize", "The size of the reference set."));
      Parameters.Add(new LookupParameter<IntValue>("NumberOfHighQualitySolutions", "The number of high quality solutions that should be added to the reference set."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The value which represents the quality of a solution."));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", "The problem's best known quality value found so far."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The value which represents the quality of a move."));
      Parameters.Add(new LookupParameter<IOperator>("SolutionCreator", "The operator which is used to create new solutions."));
      Parameters.Add(new LookupParameter<IOperator>("Evaluator", "The operator which is used to evaluate new solutions. This operator is executed in parallel, if an engine is used which supports parallelization."));
      Parameters.Add(new LookupParameter<IOperator>("Improver", "The operator which is used to improve new solutions. This operator is executed in parallelm, if an engine is used which supports parallelization."));
      Parameters.Add(new LookupParameter<VariableCollection>("Results", "The variable collection where results should be stored."));
      Parameters.Add(new LookupParameter<IOperator>("MoveGenerator", "The operator that generates the moves."));
      Parameters.Add(new LookupParameter<IOperator>("MoveMaker", "The operator that performs a move and updates the quality."));
      Parameters.Add(new LookupParameter<IOperator>("MoveEvaluator", "The operator that evaluates a move."));
      Parameters.Add(new LookupParameter<IMultiAnalyzer>("Analyzer", "The operator used to analyze the solution and moves."));
      #endregion

      #region Create operators
      Assigner assigner1 = new Assigner();
      Assigner assigner2 = new Assigner();
      ChildrenCreator childrenCreator = new ChildrenCreator();
      Comparator iterationsChecker = new Comparator();
      ConditionalBranch newSolutionsBranch = new ConditionalBranch();
      ConditionalBranch terminateBranch = new ConditionalBranch();
      IntCounter interationsCounter = new IntCounter();
      LocalSearchImprovementOperator localImprovementOperator = new LocalSearchImprovementOperator();
      Placeholder analyzer = new Placeholder();
      Placeholder solutionImprover = new Placeholder();
      Placeholder solutionEvaluator = new Placeholder();
      PopulationRebuildMethod populationRebuildMethod = new PopulationRebuildMethod();
      ResultsCollector resultsCollector = new ResultsCollector();
      ReferenceSetUpdateMethod referenceSetUpdateMethod = new ReferenceSetUpdateMethod();
      SolutionCombinationMethod solutionCombinationMethod = new SolutionCombinationMethod();
      SolutionPoolUpdateMethod solutionPoolUpdateMethod = new SolutionPoolUpdateMethod();
      SolutionsCreator solutionsCreator = new SolutionsCreator();
      SubScopesProcessor subScopesProcessor1 = new SubScopesProcessor();
      SubScopesProcessor subScopesProcessor2 = new SubScopesProcessor();
      SubScopesProcessor subScopesProcessor3 = new SubScopesProcessor();
      UniformSubScopesProcessor uniformSubScopesProcessor1 = new UniformSubScopesProcessor();
      UniformSubScopesProcessor uniformSubScopesProcessor2 = new UniformSubScopesProcessor();
      UniformSubScopesProcessor uniformSubScopesProcessor3 = new UniformSubScopesProcessor();
      #endregion

      #region Create operator graph
      OperatorGraph.InitialOperator = interationsCounter;
      assigner1.Name = "NewSolutions = true";
      assigner1.LeftSideParameter.ActualName = "NewSolutions";
      assigner1.RightSideParameter.Value = new BoolValue(true);
      assigner1.Successor = newSolutionsBranch;

      assigner2.Name = "NewSolutions = false";
      assigner2.LeftSideParameter.ActualName = "NewSolutions";
      assigner2.RightSideParameter.Value = new BoolValue(false);
      assigner2.Successor = uniformSubScopesProcessor1;

      childrenCreator.Name = "SubsetGenerator";
      childrenCreator.ParentsPerChildParameter.Value = new IntValue(2);
      childrenCreator.Successor = assigner2;

      iterationsChecker.Name = "IterationsChecker";
      iterationsChecker.Comparison.Value = ComparisonType.GreaterOrEqual;
      iterationsChecker.LeftSideParameter.ActualName = "Iterations";
      iterationsChecker.RightSideParameter.ActualName = "MaximumIterations";
      iterationsChecker.ResultParameter.ActualName = "Terminate";
      iterationsChecker.Successor = terminateBranch;

      localImprovementOperator.Name = "ImprovementMethod";
      localImprovementOperator.MaximumIterationsParameter.ActualName = MaximumIterationsParameter.Name;
      localImprovementOperator.SampleSizeParameter.ActualName = SampleSizeParameter.Name;
      localImprovementOperator.Successor = null;

      newSolutionsBranch.Name = "NewSolutionChecker";
      newSolutionsBranch.ConditionParameter.ActualName = "NewSolutions";
      newSolutionsBranch.TrueBranch = subScopesProcessor1;
      newSolutionsBranch.FalseBranch = populationRebuildMethod;
      newSolutionsBranch.Successor = null;

      terminateBranch.Name = "TerminateChecker";
      terminateBranch.ConditionParameter.ActualName = "Terminate";
      terminateBranch.TrueBranch = new EmptyOperator();
      terminateBranch.FalseBranch = referenceSetUpdateMethod;
      terminateBranch.Successor = null;

      interationsCounter.Name = "IterationCounter";
      interationsCounter.IncrementParameter.Value = new IntValue(1);
      interationsCounter.ValueParameter.ActualName = "Iterations";
      interationsCounter.Successor = resultsCollector;

      analyzer.Name = "Analyzer";
      analyzer.OperatorParameter.ActualName = "Analyzer";

      solutionImprover.Name = "SolutionImprover";
      solutionImprover.OperatorParameter.ActualName = "LocalImprovementOperator";
      solutionImprover.Successor = solutionEvaluator;

      solutionEvaluator.Name = "SolutionEvaluator";
      solutionEvaluator.OperatorParameter.ActualName = "Evaluator";
      solutionEvaluator.Successor = null;

      populationRebuildMethod.Successor = subScopesProcessor3;

      resultsCollector.CopyValue = new BoolValue(false);
      resultsCollector.CollectedValues.Add(new LookupParameter<IntValue>(IterationsParameter.Name));
      resultsCollector.CollectedValues.Add(new LookupParameter<DoubleValue>(BestKnownQualityParameter.Name, null, BestKnownQualityParameter.Name));
      resultsCollector.ResultsParameter.ActualName = ResultsParameter.Name;
      resultsCollector.Successor = iterationsChecker;

      referenceSetUpdateMethod.Successor = assigner1;

      solutionCombinationMethod.Successor = subScopesProcessor2;

      solutionPoolUpdateMethod.Successor = analyzer;

      solutionsCreator.Name = "DiversificationGenerationMethod";
      solutionsCreator.NumberOfSolutionsParameter.ActualName = "PopulationSize";
      solutionsCreator.Successor = uniformSubScopesProcessor3;

      subScopesProcessor1.DepthParameter.Value = new IntValue(1);
      subScopesProcessor1.Operators.Add(new EmptyOperator());
      subScopesProcessor1.Operators.Add(childrenCreator);
      subScopesProcessor1.Successor = newSolutionsBranch;

      subScopesProcessor2.DepthParameter.Value = new IntValue(1);
      subScopesProcessor2.Operators.Add(new EmptyOperator());
      subScopesProcessor2.Operators.Add(new EmptyOperator());
      subScopesProcessor2.Operators.Add(uniformSubScopesProcessor2);
      subScopesProcessor2.Successor = null;

      subScopesProcessor3.DepthParameter.Value = new IntValue(1);
      subScopesProcessor3.Operators.Add(solutionsCreator);
      subScopesProcessor3.Operators.Add(new EmptyOperator());
      subScopesProcessor3.Successor = interationsCounter;

      uniformSubScopesProcessor1.DepthParameter.Value = new IntValue(1);
      uniformSubScopesProcessor1.Operator = solutionCombinationMethod;
      uniformSubScopesProcessor1.Successor = solutionPoolUpdateMethod;

      uniformSubScopesProcessor2.DepthParameter.Value = new IntValue(1);
      uniformSubScopesProcessor2.Operator = solutionImprover;
      uniformSubScopesProcessor2.Successor = null;

      uniformSubScopesProcessor3.DepthParameter.Value = new IntValue(1);
      uniformSubScopesProcessor3.Operator = localImprovementOperator;
      uniformSubScopesProcessor3.Successor = null;
      #endregion
    }

    public override IOperation Apply() {
      if (MoveGeneratorParameter.ActualValue == null || MoveEvaluatorParameter.ActualValue == null || MoveMakerParameter.ActualValue == null)
        return null;
      return base.Apply();
    }
  }
}
